import math
from typing import Literal

from ovld import ovld


@ovld
def div(x: int, y: int):
    return x / y


@ovld
def div(x: str, y: str):
    return f"{x}/{y}"


@ovld
def div(x: int, y: Literal[0]):
    return math.inf


assert div(8, 2) == 4
assert div("/home", "user") == "/home/user"
assert div(10, 0) == math.inf



from typing import Callable

from ovld.core import ovld

# from plum import dispatch as ovld
# from multimethod import multimethod as ovld


class Animal:
    pass


class Mammal(Animal):
    pass


class Cat(Mammal):
    pass


@ovld
def f(t: Callable[[Mammal], Mammal]):
    return 1


@ovld
def f(t: Callable[[Mammal, Mammal], Mammal]):
    return 2


def f1(x: Mammal, y: Mammal) -> Mammal:
    return "X"


def f2(x: Animal, y: Animal) -> Mammal:
    return "X"


print(f(f1))
print(f(f2))


# @ovld
# def f(t: tuple[()]):
#     return 0

# @ovld
# def f(t: tuple[int]):
#     return 1

# @ovld
# def f(t: tuple[str]):
#     return 2

# @ovld
# def f(t: tuple[int, str]):
#     return 3

# assert f(()) == 0
# assert f((1,)) == 1
# assert f(("x",)) == 2
# assert f((2, "y")) == 3


# @ovld
# def f(d: Sequence[int]):
#     return 0


# @ovld
# def f(d: Sequence[str]):
#     return 1


# print(f([1]))
# print(f(["x"]))
# print(f([1, "x"]))
# print(f(["x", "y", "z", 1]))
# print(f([]))



import timeit
from typing import Literal
from beartype import BeartypeConf                              # <-- this isn't your fault
from beartype.claw import beartype_all, beartype_this_package  # <-- you didn't sign up for this
from beartype import beartype
from ovld import ovld
from strongtyping.strong_typing import match_typing

beartype_all(conf=BeartypeConf())
# beartype_this_package()                                        # <-- raise exceptions in your code
# beartype_all(conf=BeartypeConf(violation_type=UserWarning))

@match_typing
def f1(x: int, y: int):
    return x + y


@ovld
def f2(x: int, y: int):
    return x + y


def f3(x: int, y: int):
    assert isinstance(x, int) and isinstance(y, int)
    return x + y


print(timeit.timeit(stmt=lambda: f1(3, 4), number=100000))
print(timeit.timeit(stmt=lambda: f2(3, 4), number=100000))
print(timeit.timeit(stmt=lambda: f3(3, 4), number=100000))



from typing import Literal

import pytest

from .common import (
    multimethod_dispatch,
    ovld_dispatch,
    plum_dispatch,
)

###############################
# multiple dispatch libraries #
###############################


def make_tup(dispatch):
    @dispatch
    def tup(t: tuple[()]):
        return 0

    @dispatch
    def tup(t: tuple[int]):
        return 1

    @dispatch
    def tup(t: tuple[Literal["x"]]):
        return 2

    @dispatch
    def tup(t: tuple[int, str]):
        return 3

    @dispatch
    def tup(t: tuple[tuple[str]]):
        return 4

    return tup


#########
# match #
#########


def tup_match(expr):
    match expr:
        case ():
            return 0
        case (int(),):
            return 1
        case (str(),):
            return 2
        case (int(), str()):
            return 3
        case ((str(),),):
            return 4


####################
# Test definitions #
####################


def make_test(fn):
    def run():
        return [fn(()), fn((1,)), fn(("x",)), fn((2, "y"))]

    @pytest.mark.benchmark(group="tup")
    def test(benchmark):
        result = benchmark(run)
        assert result == [0, 1, 2, 3]

    return test


test_tup_ovld = make_test(make_tup(ovld_dispatch))
test_tup_plum = make_test(make_tup(plum_dispatch))
test_tup_multimethod = make_test(make_tup(multimethod_dispatch))

test_tup_custom_match = make_test(tup_match)



from ovld.core import ovld
from ovld.dependent import Equals
from ovld.types import Dataclass, Exactly, HasMethod

# U = Union[str, int]
# U = str | Equals[0]
# print(isinstance("ah", U))
# print(isinstance(0, U))
# print(is_dependent(U))


@ovld
def f(x: str | Equals[0]):
    return x


breakpoint()
print(f(0))
print(f(1))
print(f("what"))


print(Equals[4].__instancecheck__.__func__)
print(HasMethod["__len__"].__instancecheck__.__func__)
print(Exactly[int].__instancecheck__.__func__)
print(Dataclass.__instancecheck__.__func__)


print(isinstance(4, Equals[4]))
print(isinstance(4, Equals[4.0]))
print(isinstance([1], HasMethod["__len__"]))
print(issubclass(type([1]), HasMethod["__len__"]))


# from multimethod import multimethod as dispatch
# from plum import dispatch
# from runtype import multidispatch as dispatch

# from ovld import ovld as dispatch


# @dispatch
# def f(x: type[int]):
#     return "ah"


# @dispatch
# def f(x: type[object]):
#     return "something else"


# @dispatch
# def f(x: type[dict[str, object]]):
#     return ("didi", x)


# print(f(int))
# print(f(str))
# print(f(dict[str, int]))
# print(f(dict[int, int]))
